#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "pairings.h"
#include "funcs.h"
#include "commons.h"
#include "stats.h"
#include "arithm.h"
#include "affine.h"

//----------------------------------------------------------------------------
//                      Pairing in characteristic > 3
//
// WARNING: All functions in this file assume the elliptic curve is given
//          by an equation of the form y^2 = x^3 + ax + b
//
// WARNING: Excepted when mentionned otherwise, all the functions in this
//          file are NOT memory safe!
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
_get_opposite_point_aff_fp(GEN Q, GEN p) {
    //
    // Returns the opposite of the point Q on E(F_p).
    //
    // Coordinates of Q and resulting point are given by t_INTs.
    // p is a t_INT.
    //
    return mkvec2(gel(Q, 1), FP_NEG(gel(Q, 2), p));
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
_get_opposite_point_aff_fpxq(GEN Q, GEN p) {
    //
    // Returns the opposite of the point Q on E(F_{p^k}).
    //
    // Coordinates of Q and resulting point are given by ZXs.
    // p is a t_INT.
    //
    return mkvec2(gel(Q, 1), FPX_NEG(gel(Q, 2), p));
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE int
_are_opposite_points_aff_fp(MAYBE_UNUSED GEN E, GEN A, GEN B, GEN p) {
    IN();
    //
    // Return true if the points A and B on E(F_p) are opposites.
    // The coordinates of A and B are given by t_INTs.
    // p is a t_INT.
    //
    // WARNING -- Supposes the curve is given by y^2 = x^3 + a.x + b
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);

    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    GEN x1 = REM(gel(A, 1), p);
    GEN y1 = REM(gel(A, 2), p);
    GEN x2 = REM(gel(B, 1), p);
    GEN y2 = REM(gel(B, 2), p);
    
    if (!equalii(x1, x2)) {
        OUT();
        return 0;
    }
    if (equalii(y1, SUBII(p, y2))) {
        OUT();
        return 1;
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE int
_are_opposite_points_aff_fpxq(MAYBE_UNUSED GEN E, GEN A, GEN B, GEN t, GEN p) {
    IN();
    //
    // Return true if the points A and B are opposites.
    // The coordinates of A and B are given by FpXQs (i.e. Z_Xs).
    // t is a FpXQ (i.e. Z_X).
    // p is a t_INT.
    //
    // WARNING -- Supposes the curve is given by y^2 = x^3 + a.x + b
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);

    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    GEN x1 = Fq_red(gel(A, 1), t, p);
    GEN y1 = Fq_red(gel(A, 2), t, p);
    
    GEN x2 = Fq_red(gel(B, 1), t, p);
    GEN y2 = Fq_red(gel(B, 2), t, p);
        
    if (!ZX_equal(x1, x2)) {
        OUT();
        return 0;
    }
    if (ZX_equal(y1, FPX_NEG(y2, p))) {
        OUT();
        return 1;
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
GEN
mul_aff_fp(pairing_data_t* pdata, GEN a, GEN P) {
    int* naf = _naf(a);
    int  i = expi(a) + 1;    
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    GEN AP = gcopy(P);
    GEN MP = _get_opposite_point_aff_fp(P, pdata->charac);
    
    while (i >= 0) {
        AP = double_aff_fp(pdata, AP);
        switch (naf[i]) {
            case 1:
                AP = add_aff_fp(pdata, AP, P);
                break;
            case -1:
                AP = add_aff_fp(pdata, AP, MP);
                break;
            default:
                break;
        }
        i--;
    }
    free(naf);
    return AP;
}
//----------------------------------------------------------------------------
GEN
add_aff_fp(pairing_data_t* pdata, GEN P1, GEN P2) {
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    if (ell_is_inf(P1)) {
        return gcopy(P2);
    }
    if (ell_is_inf(P2)) {
        return gcopy(P1);
    }
    GEN x1 = gel(P1, 1);
    GEN y1 = gel(P1, 2);
    GEN x2 = gel(P2, 1);
    GEN y2 = gel(P2, 2);
    
    if (equalii(x1, x2)) {
        if (equalii(y1, y2)) {
            return double_aff_fp(pdata, P1);
        } else {
            return ellinf();
        }
    }
    GEN p = pdata->charac;
        
    GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);
    GEN x3    = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
    GEN y3    = FP_NEG(ADDII(y1, MULII(slope, SUBII(x3, x1))), p);

    return mkvec2(x3, y3);
}
//----------------------------------------------------------------------------
GEN
double_aff_fp(pairing_data_t* pdata, GEN P1) {
    //
    // Given P1 = (x1, y1) then [2]P1 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x1
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (3.x1^2 + a) / (2y1)
    //
    if (ell_is_inf(P1)) {
        return ellinf();
    }
    GEN x1 = gel(P1, 1);
    GEN y1 = gel(P1, 2);
    if (cmpis(y1, 0) == 0) {
        return ellinf();
    }
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    
    GEN _2_x1    = MULUI(2, x1);
    GEN _2_y1    = MULUI(2, y1);
    GEN _sqrx1   = SQRI(x1);
    GEN _3_sqrx1 = MULUI(3, _sqrx1);
    GEN slope    = FP_DIV(ADDII(_3_sqrx1, gel(gel(E, 4), 2)), _2_y1, p);
    
    GEN x3 = FP_SUB(SQRI(slope), _2_x1, p);
    GEN y3 = FP_NEG(ADDII(y1, MULII(slope, SUBII(x3, x1))), p);

    return mkvec2(x3, y3);
}
//----------------------------------------------------------------------------
GEN
naf_mul_aff_fpxq(pairing_data_t* pdata, GEN a, GEN Q) {
    int* naf = _naf(a);
    int  i = expi(a) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    GEN AQ = gcopy(Q);
    GEN MQ = _get_opposite_point_aff_fpxq(Q, pdata->charac);
    
    while (i >= 0) {
        AQ = double_aff_fpxq(pdata, AQ);
        switch (naf[i]) {
            case 1:
                AQ = add_aff_fpxq(pdata, AQ, Q);
                break;
            case -1:
                AQ = add_aff_fpxq(pdata, AQ, MQ);
                break;
            default:
                break;
        }
        i--;
    }
    free(naf);

    return AQ;
}
//----------------------------------------------------------------------------
GEN
std_mul_aff_fpxq(pairing_data_t* pdata, GEN a, GEN Q) {
    long i = expi(a) - 1;
    
    GEN AQ = gcopy(Q);
    while (i >= 0) {
        AQ = double_aff_fpxq(pdata, AQ);
        if (_get_bit(a, i)) {
            AQ = add_aff_fpxq(pdata, AQ, Q);
        }
        i--;
    }
    return AQ;
}
//----------------------------------------------------------------------------
GEN
mul_aff_fpxq(pairing_data_t* pdata, GEN a, GEN Q) {
    return naf_mul_aff_fpxq(pdata, a, Q);
}
//----------------------------------------------------------------------------
GEN
add_aff_fpxq(pairing_data_t* pdata, GEN Q1, GEN Q2) {
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P2 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    if (ell_is_inf(Q1)) {
        return gcopy(Q2);
    }
    if (ell_is_inf(Q2)) {
        return gcopy(Q1);
    }
    GEN x1 = gel(Q1, 1);
    GEN y1 = gel(Q1, 2);
    GEN x2 = gel(Q2, 1);
    GEN y2 = gel(Q2, 2);
    
    if (_ZX_equal(x1, x2)) {
        if (_ZX_equal(y1, y2)) {
            return double_aff_fpxq(pdata, Q1);
        } else {
            return ellinf();
        }
    }
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), t, p);
    
    GEN x3 = FPX_SUB(ZX_SUB(ZX_SQR(slope), x1), x2, p);
        x3 = FPX_REM(FpX_red(x3, p), t, p);
    GEN y3 = FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p);
        y3 = FPX_REM(FpX_red(y3, p), t, p);
    
    return mkvec2(x3, y3);
}
//----------------------------------------------------------------------------
GEN
double_aff_fpxq(pairing_data_t* pdata, GEN Q1) {
    //
    // Given Q1 = (x1, y1) then [2]Q1 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x1
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (3.x1^2 + a) / (2.y1)
    //
    if (ell_is_inf(Q1)) {
        return ellinf();
    }
    GEN x1 = gel(Q1, 1);
    GEN y1 = gel(Q1, 2);
    if (cmpis(y1, 0) == 0) {
        return ellinf();
    }
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN _2_x1    = ZX_MULS(x1, 2);
    GEN _2_y1    = ZX_MULS(y1, 2);
    GEN _sqrx1   = ZX_SQR(x1);
    GEN _3_sqrx1 = ZX_MULS(_sqrx1, 3);

    GEN slope = FPXQ_DIV(ZX_Z_ADD(_3_sqrx1, gel(gel(E, 4), 2)), _2_y1, t, p);

    GEN x3 = FPX_SUB(ZX_SQR(slope), _2_x1, p);
        x3 = FPX_REM(FpX_red(x3, p), t, p);
    GEN y3 = FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p);
        y3 = FPX_REM(FpX_red(y3, p), t, p);
    
    return mkvec2(x3, y3);
}
//----------------------------------------------------------------------------
GEN
eval_vert_aff_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // A is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // Return value is a ZX.
    //
    if (ell_is_inf(A)) {
        OUT();
        return Rg_to_FpXQ(gen_1, pdata->irredpol, pdata->charac);
    }
    OUT();
    
    return FPX_FP_ADD(
             gel(Q, 1), FP_NEG(gel(A, 1), pdata->charac), pdata->charac
           );
}
//----------------------------------------------------------------------------
void
update_vert_aff_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // A is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // Return value is a ZX.
    //
    if (pdata->do_denom_elim) {
        return;
    }
    if (!ell_is_inf(A)) {
        GEN p = pdata->charac;
        
        pdata->denom = FPXQ_MUL(
                           pdata->denom,
                           FPX_FP_ADD(gel(Q, 1), FP_NEG(gel(A, 1), p), p),
                           pdata->irredpol,
                           p
                       );
        OUT();
        return;
    }
}
//----------------------------------------------------------------------------
void
update_tangent_aff_fp_fpxq(pairing_data_t* pdata, GEN Q)
{
    IN();
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point Q to update pf->num and finally doubles the point
    // pf->point.
    //
    // pf.point is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) then [2]P2 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x
    //     y3 = -(slope * (x3 - x) + y)
    //
    // with: slope = (3.x^2 + a) / (2y)
    //
    // The line tangent to [2]P has for equation:
    //
    //     Y - slope * (X - x) - y
    //
    if (ell_is_inf(pdata->point)) {
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    }
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    if (cmpis(y1, 0) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        GEN vert = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        pdata->num = FPXQ_MUL(pdata->num, vert, t, p);
        pdata->point = ellinf();
        OUT();
        return;
    }

    GEN _2_x1    = MULUI(2, x1);
    GEN _2_y1    = MULUI(2, y1);
    GEN _sqrx1   = SQRI(x1);
        _sqrx1   = REM(_sqrx1, p);
    GEN _3_sqrx1 = MULUI(3, _sqrx1);

    GEN slope = FP_DIV(ADDII(_3_sqrx1, gel(gel(E, 4), 2)), _2_y1, p);

    GEN x3 = FP_SUB(SQRI(slope), _2_x1, p);
    GEN y3 = ADDII(y1, MULII(slope, SUBII(x3, x1)));
    y3 = FP_NEG(y3, p);

    GEN tmp1 = FPX_FP_ADD(gel(Q, 1), FP_NEG(x1, p), p);
    GEN tmp2 = FPX_FP_ADD(FPX_FP_MUL(tmp1, slope, p), y1, p);
    GEN line = FPX_SUB(gel(Q, 2), tmp2, p);

    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;

    if (!pdata->do_denom_elim) {
        GEN vertical = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        pdata->denom = FPXQ_MUL(pdata->denom, vertical, t, p);
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_inv_tangent_aff_fp_fpxq(pairing_data_t* pdata, GEN Q)
{
    IN();
    //
    // Computes the equation of the line tangent to -(pdata.point), evaluates
    // it at the point Q to update pf->denom (indeed, denom!) and finally
    // doubles the point pf->point.
    //
    // pdata.point is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x, y) then [2]P2 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x
    //     y3 = -(slope * (x3 - x) + y)
    //
    // with: slope = (3.x^2 + a) / (2y)
    //
    // The line tangent to [2]P has for equation:
    //
    //     Y - slope * (X - x) - y
    //
    if (ell_is_inf(pdata->point)) {
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    } 
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    //
    // (x1, y1) are the coordinates of -(pdata->point)
    //
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = FP_NEG(gel(pdata->point, 2), p);

    if (cmpis(y1, 0) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        GEN line     = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);        
        pdata->point = ellinf();
        OUT();
        return;
    }
    GEN _2_x1    = MULUI(2, x1);
    GEN _2_y1    = MULUI(2, y1);
    GEN _sqrx1   = SQRI(x1);
        _sqrx1   = REM(_sqrx1, p);
    GEN _3_sqrx1 = MULUI(3, _sqrx1);

    GEN slope = FP_DIV(ADDII(_3_sqrx1, gel(gel(E, 4), 2)), _2_y1, p);
    
    //
    // (x3, y3) are the coordinates of the point 2 * pdata->point
    //
    GEN x3 = FP_SUB(SQRI(slope), _2_x1, p);
    GEN y3 = REM(ADDII(y1, MULII(slope, SUBII(x3, x1))), p);

    GEN tmp1 = FPX_FP_ADD(gel(Q, 1), FP_NEG(x1, p), p);
    GEN tmp2 = FPX_FP_ADD(FPX_FP_MUL(tmp1, slope, p), y1, p);
    GEN line = FPX_SUB(gel(Q, 2), tmp2, p);

    pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);
    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_aff_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    //
    // Computes the equation of the line through pf.point and A (supposed to
    // be different from the point at infinity), evaluates it at the point Q
    // to update pf->num, adds A to the point pfunc->point and finally
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point Q.
    //
    // pf.point and A are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    // The line (P1-P2) has for equation:
    //
    //     Y - slope * (X - x1) - y1
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (equalii(gel(A, 1), gel(T, 1))) {
        if (t_is_inf || equalii(gel(A, 2), SUBII(p, gel(T, 2)))) {
            //
            // In other words T and A are on the same vertical line.
            //
            GEN line     = eval_vert_aff_fp_fpxq(pdata, T, Q);
            pdata->num   = FPXQ_MUL(pdata->num, line, t, p);
            pdata->point = ellinf();
            OUT();
            return;

        } else if (equalii(gel(A, 2), gel(T, 2))) {
            //
            // The two points T and A are the same. Compute the tangent at T.
            //
            update_tangent_aff_fp_fpxq(pdata, Q);
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(A, 1);
    GEN y2 = gel(A, 2);

    GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);

    GEN x3 = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
    GEN y3 = ADDII(y1, MULII(slope, SUBII(x3, x1)));
    y3 = FP_NEG(y3, p);
    
    GEN tmp1 = FPX_FP_MUL(FPX_FP_ADD(gel(Q, 1), FP_NEG(x1, p), p), slope, p);
    GEN tmp2 = FPX_FP_ADD(tmp1, y1, p);
    GEN line = FPX_SUB(gel(Q, 2), tmp2, p);

    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;

    if (!pdata->do_denom_elim) {
        GEN vertical = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        pdata->denom = FPXQ_MUL(pdata->denom, vertical, t, p);
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_naf_aff_fp_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Used to compute f_{n,P}(Q) whenever we have a -1 in the NAF
    // decomposition of n. In that case we have to divide the current
    // f_{n',P}(Q) value by v_P(Q) which actually simplify the computation
    // of the line. Hence this function.
    //
    // It computes the equation of the line through pf.point and [-1]P (supposed
    // to be different from the point at infinity), evaluates it at the point Q
    // to update pf->num, adds [-1]P to the point pf->point,
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point Q and multiplies pf->denom by the
    // vertical line at P evaluated at the point Q.
    //
    // pf.point and P are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P = (x1, y1) and P2 = (x2, y2) then [-1]P + P2 = (x3, y3)
    // is given by:
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) - y1)
    //
    // with: slope = (y2 + y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (equalii(gel(P, 1), gel(T, 1))) {
        if (t_is_inf || equalii(gel(P, 2), gel(T, 2))) {
            //
            // In other words T and [-1]P are on the same vertical line.
            //
            pdata->point = ellinf();
            OUT();
            return;

        } else if (equalii(gel(P, 2), SUBII(p, gel(T, 2)))) {
            //
            // The two points T and [-1]P are the same. Compute tangent at T.
            //
            update_tangent_aff_fp_fpxq(pdata, Q);
            pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(P, 1);
    GEN y2 = FP_NEG(gel(P, 2), p);

    GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);

    GEN x3 = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
    GEN y3 = ADDII(y1, MULII(slope, SUBII(x3, x1)));
    y3 = FP_NEG(y3, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;

    //
    // Note pf.point = T. The line ([-1]P.T) evaluated at Q is given by:
    //     yQ - slope * (xQ - xP) + yP with slope = (yT + yP) / (xT - xP)
    // And once divided by v_{Q}(P) we simply obtain:
    //     (yQ + yP) / (xQ - xP) - slope
    // where (yQ + yP) / (xQ - xP) was precomputed and stored in pdata->tmp.
    //    
    GEN line   = FPX_FP_ADD(pdata->tmp, FP_NEG(slope, p), p);
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);
    
    if (!pdata->do_denom_elim) {
        GEN vertical = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        pdata->denom = FPXQ_MUL(pdata->denom, vertical, t, p);
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_inv_if1_mm_chord_aff_fp_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Computes the equation of the line through [-1]pf.point and [-1]P
    // (supposed to be different from the point at infinity), evaluates it at
    // the point Q to update pf->denom (indeed denom, not num!), adds P to
    // the point pf->point, and multiplies pf->denom by the vertical line
    // at P evaluated at the point Q.
    //
    // pf.point and P are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (equalii(gel(P, 1), gel(T, 1))) {
        if (t_is_inf || equalii(gel(P, 2), SUBII(p, gel(T, 2)))) {
            //
            // In other words T = [-]P
            //
            //pariprintf("%s: ---- WARNING ---- exit branch #0\n", __func__);
            pdata->point = ellinf();
            // ---- del next line for a correct result (has been tested)
            //pdata->denom = FPXQ_MUL(pdata->denom, pdata->tmp2, t, p);
            //
            OUT();
            return;

        } else if (equalii(gel(P, 2), gel(T, 2))) {
            //
            // The two points [-1]T and [-1]P are the same. Compute the
            // tangent at [-T].
            //
            //pariprintf("%s: ---- WARNING ---- Untested branch\n", __func__);
            update_boxalletal_inv_tangent_aff_fp_fpxq(pdata, Q);
            pdata->denom = FPXQ_MUL(pdata->denom, pdata->tmp2, t, p);
            //
            // Recall that pdata->tmp2 is 1/v_P(Q)
            //
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(P, 1);
    GEN y2 = gel(P, 2);
    
    GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);

    GEN x3 = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
    GEN y3 = ADDII(y1, MULII(slope, SUBII(x3, x1)));
    y3 = FP_NEG(y3, p);
    //
    // Note pf.point = T. The line ([-1]P.[-1]T) evaluated at Q is given by:
    //     yQ - slope * (xQ - xP) + yP with slope = - (yT - yP) / (xT - xP)
    // And once divided by v_{Q}(P) we simply obtain:
    //     (yQ + yP) / (xQ - xP) - slope
    // where (yQ + yP) / (xQ - xP) was precomputed and stored in pdata->tmp.
    //
    GEN line     = FPX_FP_ADD(pdata->tmp, slope, p);
    pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_inv_mp_chord_aff_fp_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Computes the equation of the line through [-1]pf.point and P
    // (supposed to be different from the point at infinity), evaluates it at
    // the point Q to update pf->denom (indeed denom, not num!), adds [-1]P to
    // the point pf->point.
    //
    // pf.point and P are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (equalii(gel(P, 1), gel(T, 1))) {
        if (t_is_inf || equalii(gel(P, 2), gel(T, 2))) {
            //
            // In other words T = P, so -T + P = 0
            //
            pdata->point = ellinf();
            //pariprintf("%s: ---- WARNING ---- exit branch #1\n", __func__);
            // is that correct? Yes, it is! (has been tested!)
            pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            //
            OUT();
            return;

        } else if (equalii(gel(P, 2), SUBII(p, gel(T, 2)))) {
            //
            // The two points [-1]T and P are the same. Compute the
            // tangent at [-T].
            //
            //pariprintf("%s: ---- WARNING ---- Untested branch\n", __func__);
            update_boxalletal_inv_tangent_aff_fp_fpxq(pdata, Q);
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(P, 1);
    GEN y2 = gel(P, 2);
    y2 = FP_NEG(y2, p);
    
    GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);

    GEN x3 = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
    GEN y3 = ADDII(y1, MULII(slope, SUBII(x3, x1)));
    y3 = FP_NEG(y3, p);
    
    GEN tmp1 = FPX_FP_MUL(FPX_FP_ADD(gel(Q, 1), FP_NEG(x2, p), p), slope, p);
    GEN tmp2 = FPX_FP_ADD(tmp1, y2, p);
    GEN line = FPX_ADD(gel(Q, 2), tmp2, p);

    pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);
    
    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_if1_pm_chord_aff_fp_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Computes the equation of the line through pf.point and [-1]P
    // (supposed to be different from the point at infinity), evaluates it at
    // the point Q to update pf->num, adds [-1]P to the point pf->point, and
    // multiplies pf->num by the vertical line at P evaluated at the point Q.
    //
    // pf.point and P are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (equalii(gel(P, 1), gel(T, 1))) {
        if (t_is_inf || equalii(gel(P, 2), gel(T, 2))) {
            //
            // In other words T - P = 0
            //
            //pariprintf("%s: ---- WARNING ---- exist branch #2\n", __func__);
            
            pdata->point = ellinf();
            //pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            OUT();
            return;

        } else if (equalii(gel(P, 2), SUBII(p, gel(T, 2)))) {
            //
            // The two points T and [-1]P are the same. Compute the
            // tangent at T.
            //
            //pariprintf("%s: ---- WARNING ---- Untested branch\n", __func__);
            update_tangent_aff_fp_fpxq(pdata, Q);
            pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(P, 1);
    GEN y2 = gel(P, 2);
    y2 = FP_NEG(y2, p);
    
    GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);

    GEN x3 = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
    GEN y3 = ADDII(y1, MULII(slope, SUBII(x3, x1)));
    y3 = FP_NEG(y3, p);
    //
    // Note pf.point = T. The line ([-1]P.T) evaluated at Q is given by:
    //     yQ - slope * (xQ - xP) + yP with slope = (yT + yP) / (xT - xP)
    // And once divided by v_{Q}(P) we simply obtain:
    //     (yQ + yP) / (xQ - xP) - slope
    // where (yQ + yP) / (xQ - xP) was precomputed and stored in pdata->tmp.
    //
    GEN line   = FPX_FP_ADD(pdata->tmp, FP_NEG(slope, p), p);
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
GEN
eval_vert_aff_fpxq_fp(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // pf.point and A are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. Q is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    if (ell_is_inf(A)) {
        OUT();
        return Rg_to_FpXQ(gen_1, pdata->irredpol, pdata->charac);
    }
    GEN p = pdata->charac;
    GEN v = FPX_FP_ADD(FPX_NEG(gel(A, 1), p), gel(Q, 1), p);
    OUT();
    return v;
}
//----------------------------------------------------------------------------
void
update_vert_aff_fpxq_fp(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // pf.point and A are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. Q is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    if (pdata->do_denom_elim) {
        return;
    }
    if (!ell_is_inf(A)) {
        GEN p = pdata->charac;
        GEN vert = FPX_FP_ADD(FPX_NEG(gel(A, 1), p), gel(Q, 1), p);
        pdata->denom = FPXQ_MUL(pdata->denom, vert, pdata->irredpol, p);
        OUT();
        return;
    }
}
//----------------------------------------------------------------------------
void
update_tangent_aff_fpxq_fp(pairing_data_t* pdata, GEN P)
{
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point P to update pf->num and finally doubles the point
    // pf->point.
    //
    // pf.point is a point on E(F_{p^k}) whose coordinates are given by 
    // ZXs. P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given P1 = (x1, y1) then [2]P2 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x
    //     y3 = -(slope * (x3 - x) + y)
    //
    // with: slope = (3.x^2 + a) / (2y)
    //
    // The line tangent to [2]P has for equation:
    //
    //     Y - slope * (X - x) - y
    //
    IN();
    
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    if (signe(y1) == 0) {
        GEN vert   = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        pdata->num = FPXQ_MUL(pdata->num, vert, t, p);
        pdata->point = ellinf();
        OUT();
        return;
    }
    GEN _2_x1    = ZX_MULS(x1, 2);
    GEN _2_y1    = ZX_MULS(y1, 2);
    GEN _sqrx1   = ZX_SQR(x1);
        _sqrx1   = FPX_REM(_sqrx1, t, p);
    GEN _3_sqrx1 = ZX_MULS(_sqrx1, 3);

    GEN slope = FPXQ_DIV(
                  ZX_Z_ADD(_3_sqrx1, gel(gel(E, 4), 2)),
                  _2_y1,
                  t,
                  p
                );

    GEN x3 = FPX_SUB(ZX_SQR(slope), _2_x1, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p);
    GEN y3 = FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p);
        y3 = FPX_REM(FPX_RED(y3, p), t, p);

    GEN tmp1 = FPX_FP_ADD(FPX_NEG(x1, p), gel(P, 1), p);
    GEN tmp2 = FPX_NEG(FPX_ADD(FPXQ_MUL(slope, tmp1, t, p), y1, p), p);
    GEN line = FPX_FP_ADD(tmp2, gel(P, 2), p);

    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;

    if (!pdata->do_denom_elim) {
        GEN vertical = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        pdata->denom = FPXQ_MUL(pdata->denom, vertical, t, p);
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_inv_tangent_aff_fpxq_fp(pairing_data_t* pdata, GEN P)
{
    IN();
    //
    // Computes the equation of the line tangent to -(pdata.point), evaluates
    // it at the point P to update pf->denom and finally doubles the point
    // pf->point.
    //
    // pdata.point is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given P1 = (x, y) then [2]P2 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x
    //     y3 = -(slope * (x3 - x) + y)
    //
    // with: slope = (3.x^2 + a) / (2y)
    //
    // The line tangent to [2]P has for equation:
    //
    //     Y - slope * (X - x) - y
    //
    if (ell_is_inf(pdata->point)) {
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    } 
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;

    //
    // (x1, y1) are the coordinates of -(pdata->point)
    //
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = FPX_NEG(gel(pdata->point, 2), p);

    if (signe(y1) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        GEN line = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);
        pdata->point = ellinf();
        OUT();
        return;
    }

    GEN _2_x1    = ZX_MULS(x1, 2);
    GEN _2_y1    = ZX_MULS(y1, 2);
    GEN _sqrx1   = ZX_SQR(x1);
        _sqrx1   = FPX_REM(_sqrx1, t, p);
    GEN _3_sqrx1 = ZX_MULS(_sqrx1, 3);

    GEN slope = FPXQ_DIV(ZX_Z_ADD(_3_sqrx1, gel(gel(E, 4), 2)), _2_y1, t, p);

    //
    // (x3, y3) are the coordinates of the point 2 * pdata->point
    //
    GEN x3 = FPX_SUB(ZX_SQR(slope), _2_x1, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p);
    GEN y3 = ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1)));
        y3 = FPX_REM(FPX_RED(y3, p), t, p);

    GEN tmp1 = FPXQ_MUL(slope, FPX_FP_ADD(FPX_NEG(x1, p), gel(P, 1), p), t, p);
    GEN tmp2 = FPX_NEG(FPX_ADD(tmp1, y1, p), p);
    GEN line = FPX_FP_ADD(tmp2, gel(P, 2), p);
    
    pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);
    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_aff_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Computes the equation of the line through pf.point and Q, evaluates
    // it at the point P to update pf->num and finally adds Q to the point
    // pfunc->point.
    //
    // pf.point and Q are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P2 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    // The line (P1-P2) has for equation:
    //
    //     Y - slope * (X - x1) - y1
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (_ZX_equal(gel(Q, 1), gel(T, 1))) {
        if (t_is_inf ||
            _ZX_equal(gel(Q, 2), FpX_neg(gel(T, 2), p))) {
            
            GEN vertical = eval_vert_aff_fpxq_fp(pdata, T, P);
            pdata->num   = FPXQ_MUL(pdata->num, vertical, t, p);
            pdata->point = ellinf();
            OUT();
            return;

        } else if (_ZX_equal(gel(Q, 2), gel(T, 2))) {
            update_tangent_aff_fpxq_fp(pdata, P);
            OUT();
            return;
        }
    }
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);

    GEN x2 = gel(Q, 1);
    GEN y2 = gel(Q, 2);

    GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), t, p);

    GEN x3 = FPX_SUB(ZX_SUB(ZX_SQR(slope), x1), x2, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p);    
    GEN y3 = FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p);
        y3 = FPX_REM(FPX_RED(y3, p), t, p);

    GEN tmp1 = FPX_FP_ADD(FPX_NEG(x1, p), gel(P, 1), p);
    GEN tmp2 = FPX_NEG(FPX_ADD(FPXQ_MUL(slope, tmp1, t, p), y1, p), p);
    GEN line = FPX_FP_ADD(tmp2, gel(P, 2), p);
    
    line = FPX_REM(line, t, p);
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;

    if (!pdata->do_denom_elim) {
        GEN vertical = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        pdata->denom = FPXQ_MUL(pdata->denom, vertical, t, p);
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_naf_aff_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Used to compute f_{n,Q}(P) whenever we have a -1 in the NAF
    // decomposition of n. In that case we have to divide the current
    // f_{n',Q}(P) value by v_Q(P) which actually simplify the computation
    // of the line. Hence this function.
    //
    // It computes the equation of the line through pf.point and [-1]Q (supposed
    // to be different from the point at infinity), evaluates it at the point P
    // to update pf->num, adds [-1]Q to the point pf->point,
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point P and multiplies pf->denom by the
    // vertical line at Q evaluated at the point P.
    //
    // P is a point on E(F_p) whose coordinates are given by t_INTs. 
    // pf.point and Q are points on E(F_{p^k}) whose coordinates are given
    // by ZXs.
    //
    // Given P = (x1, y1) and P2 = (x2, y2) then [-1]P + P2 = (x3, y3)
    // is given by:
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) - y1)
    //
    // with: slope = (y2 + y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (_ZX_equal(gel(Q, 1), gel(T, 1))) {
        if (t_is_inf || _ZX_equal(gel(Q, 2), gel(T, 2))) {
            pdata->point = ellinf();
            OUT();
            return;
        } else if (_ZX_equal(gel(Q, 2), FpX_neg(gel(T, 2), p))) {
            update_tangent_aff_fpxq_fp(pdata, P);
            pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            OUT();
            return;
        }
    }
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);

    GEN x2 = gel(Q, 1);
    GEN y2 = FPX_NEG(gel(Q, 2), p);

    GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), t, p);

    GEN x3 = FPX_SUB(ZX_SUB(ZX_SQR(slope), x1), x2, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p);
    GEN y3 = FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p);
        y3 = FPX_REM(FPX_RED(y3, p), t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;

    //
    // Note pf.point = T. The line ([-1]Q.T) evaluated at P is given by:
    //     yP - slope * (xP - xQ) + yQ with slope = (yT + yQ) / (xT - xQ)
    // And once divided by v_{P}(Q) we simply obtain:
    //     (yQ + yP) / (xP - xQ) - slope
    // where (yQ + yP) / (xP - xQ) was precomputed and stored in pdata->tmp.
    //
    GEN line   = FPX_ADD(pdata->tmp, FPX_NEG(slope, p), p);
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);
        
    if (!pdata->do_denom_elim) {
        GEN vertical = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        pdata->denom = FPXQ_MUL(pdata->denom, vertical, t, p);
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_inv_if1_mm_chord_aff_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Computes the equation of the line through [-1]pf.point and [-1]Q
    // (supposed to be different from the point at infinity), evaluates it at
    // the point P to update pf->denom (indeed denom, not num!), adds Q to
    // the point pf->point, multiplies pf->num (indeed num!) by the equation
    // of the vertical line at pf->point evaluated at the point P and
    // multiplies pf->num by the vertical line at Q evaluated at the point P.
    //
    // pf.point and Q are points on E(F_p^k) whose coordinates are given by 
    // ZXs. P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);
    
    if (_ZX_equal(gel(Q, 1), gel(T, 1))) {
        if (t_is_inf || _ZX_equal(gel(Q, 2), FpX_neg(gel(T, 2), p))) {
            //
            // In other words T and A are on the same vertical line.
            //
            pdata->point = ellinf();
            OUT();
            return;

        } else if (_ZX_equal(gel(Q, 2), gel(T, 2))) {
            //
            // The two points T and P are the same. Compute the tangent at T.
            //
            update_boxalletal_inv_tangent_aff_fpxq_fp(pdata, P);
            pdata->denom = FPXQ_MUL(pdata->denom, pdata->tmp2, t, p);
            OUT();
            return;
        }
    }
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);

    GEN x2 = gel(Q, 1);
    GEN y2 = gel(Q, 2);

    GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), t, p);
    
    GEN x3 = FPX_SUB(ZX_SUB(ZX_SQR(slope), x1), x2, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p);
    GEN y3 = FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p);
        y3 = FPX_REM(FPX_RED(y3, p), t, p);
    
    //
    // Note pf.point = T. The line ([-1]Q.[-1]T) evaluated at P is given by:
    //     yP - slope * (xP - xQ) + yQ with slope = - (yT - yQ) / (xT - xQ)
    // And once divided by v_{P}(Q) we simply obtain:
    //     (yP + yQ) / (xP - xQ) - slope
    // where (xP + yQ) / (xP - xQ) was precomputed and stored in pdata->tmp.
    //
    GEN line     = FPX_ADD(pdata->tmp, slope, p);
    pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_inv_mp_chord_aff_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Computes the equation of the line through [-1]pf.point and Q
    // (supposed to be different from the point at infinity), evaluates it at
    // the point P to update pf->denom (indeed denom, not num!), adds [-1]Q to
    // the point pf->point.
    //
    // P is a point on E(F_p) whose coordinates are given by t_INTs.
    // pf.point and Q are points on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (_ZX_equal(gel(Q, 1), gel(T, 1))) {
        if (t_is_inf || _ZX_equal(gel(Q, 2), gel(T, 2))) {
            //
            // In other words T = Q, so -T + Q = 0
            //
            pdata->point = ellinf();
            //pariprintf("%s: ---- WARNING ---- exist branch #1\n", __func__);
            // is that correct? I don't know!
            pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            //
            OUT();
            return;

        } else if (_ZX_equal(gel(Q, 2), FpX_neg(gel(T, 2), p))) {
            //
            // The two points [-1]T and P are the same. Compute the
            // tangent at [-T].
            //
            //pariprintf("%s: ---- WARNING ---- Untested branch\n", __func__);
            update_boxalletal_inv_tangent_aff_fpxq_fp(pdata, P);
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(Q, 1);
    GEN y2 = gel(Q, 2);
    y2 = FPX_NEG(y2, p);
        
    GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), t, p);

    GEN x3 = FPX_SUB(ZX_SUB(ZX_SQR(slope), x1), x2, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p);        
    GEN y3 = FPX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1)), p);
        y3 = FPX_NEG(y3, p);
        y3 = FPX_REM(FPX_RED(y3, p), t, p);
    
    GEN tmp1 = FPXQ_MUL(FPX_FP_ADD(FPX_NEG(x2, p), gel(P, 1), p), slope, t, p);
    GEN tmp2 = FPX_ADD(tmp1, y2, p);
    GEN line = FPX_FP_ADD(tmp2, gel(P, 2), p);

    line = FPX_REM(line, t, p);
    
    pdata->denom = FPXQ_MUL(pdata->denom, line, t, p);
    
    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------
void
update_boxalletal_if1_pm_chord_aff_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Computes the equation of the line through pf.point and [-1]Q
    // (supposed to be different from the point at infinity), evaluates it at
    // the point P to update pf->num, adds [-1]Q to the point pf->point, and
    // multiplies pf->num by the vertical line at Q evaluated at the point P.
    //
    // P is a point on E(F_p) whose coordinates are given by t_INTs.
    // pf.point and Q are points on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P1 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    int t_is_inf = ell_is_inf(T);

    if (_ZX_equal(gel(Q, 1), gel(T, 1))) {
        if (t_is_inf || _ZX_equal(gel(Q, 2), gel(T, 2))) {
            //
            // In other words T - Q = 0
            //
            //pariprintf("%s: ---- WARNING ---- exist branch #2\n", __func__);
            
            pdata->point = ellinf();
            OUT();
            return;

        } else if (equalii(gel(Q, 2), FpX_neg(gel(T, 2), p))) {
            //
            // The two points T and [-1]P are the same. Compute the
            // tangent at T.
            //
            //pariprintf("%s: ---- WARNING ---- Untested branch\n", __func__);
            update_tangent_aff_fpxq_fp(pdata, P);
            pdata->num = FPXQ_MUL(pdata->num, pdata->tmp2, t, p);
            OUT();
            return;
        }
    }
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);

    GEN x2 = gel(Q, 1);
    GEN y2 = gel(Q, 2);
    y2 = FPX_NEG(y2, p);
       
    GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), t, p);

    GEN x3 = FPX_SUB(ZX_SUB(FPXQ_SQR(slope, t, p), x1), x2, p);
        x3 = FPX_REM(FPX_RED(x3, p), t, p); 
    
    GEN y3 = FPX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1)), p);
        y3 = FPX_NEG(y3, p);
        y3 = FPX_REM(FPX_RED(y3, p), t, p);

    GEN line   = FPXQ_ADD(pdata->tmp, FPX_NEG(slope, p), t, p);
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = x3;
    gel(pdata->point, 2) = y3;
    OUT();
}
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Miller functions
//----------------------------------------------------------------------------
MAYBE_UNUSED GEN
std_miller_aff_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Standard Miller algorithm computing f_{r,P}(Q).
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    long i = expi(absr) - 1;
    
    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_aff_fp_fpxq(pdata, Q);

        if (_get_bit(absr, i)) {
            update_chord_aff_fp_fpxq(pdata, P, Q);
        }
        i--;
    }
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        pdata->num = FPXQ_MUL(pdata->num, vrp, T, p);
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
        OUT();
        if (pdata->do_denom_elim) {
            return FPXQ_INV(pdata->num, T, p);
        }
        return FPXQ_DIV(pdata->denom, pdata->num, T, p);
    }
    OUT();
    if (pdata->do_denom_elim) {
        return pdata->num;
    }
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
naf_miller_aff_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();    
    //
    // Miller algorithm computing f_{r,P}(Q) using a NAF representation of r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));

    GEN p = pdata->charac;
    GEN T = pdata->irredpol;

    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);
    
    int  i = expi(absr) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    pdata->tmp2 = FPXQ_INV(eval_vert_aff_fp_fpxq(pdata, P, Q), T, p);
    pdata->tmp  = FPXQ_MUL(pdata->tmp2, ZX_Z_ADD(gel(Q, 2), gel(P, 2)), T, p);
    
    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_aff_fp_fpxq(pdata, Q);

        if (naf[i] == 1) {
            update_chord_aff_fp_fpxq(pdata, P, Q);
        } else {
            if (naf[i] == -1) {
                //
                // f_{i-1, P} = f_{i, P} * l_{iP, -P} / (v_{P} * v_{(i-1)P})
                // but since l_{iP, -P} / v_{P} can be more easily computed,
                // use a tuned function that does just that...
                //
                update_chord_naf_aff_fp_fpxq(pdata, P, Q);
            }
        }
        i--;
    }
    free(naf);
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(rQ))
        //
        GEN vrp = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;        
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED GEN
boxalletal_miller_aff_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Miller variant computing f_{r,P}(Q) from the paper "A variant of
    // Miller's formula and algorithm by Boxall, El Mrabet, Laguillaumie
    // and Le.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    unsigned short old_do_denom_elim = pdata->do_denom_elim;
    pdata->do_denom_elim = 1;
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    
    long l = expi(absr) + 1;
    long h = 0;
    for (long j = 0; j < l; j ++) {
        h += _get_bit(absr, j);
    }
    int delta = 0;
    pdata->tmp2 = FPXQ_INV(eval_vert_aff_fp_fpxq(pdata, P, Q), T, p);

    if (IS_ODD(l + h)) {
        delta = 1;
        pdata->denom = gcopy(pdata->tmp2);
    } else {
        delta = 0;
    }
    pdata->tmp = FPXQ_MUL(pdata->tmp2, ZX_Z_ADD(gel(Q, 2), gel(P, 2)), T, p);
    
    long i = l - 2;
    while (i >= 0) {
        if (delta == 0) {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_tangent_aff_fp_fpxq(pdata, Q);
            delta = 1;
            if (_get_bit(absr, i)) {
                update_boxalletal_inv_if1_mm_chord_aff_fp_fpxq(pdata, P, Q);
                delta = 0;
            }
        
        } else {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_boxalletal_inv_tangent_aff_fp_fpxq(pdata, Q);
            delta = 0;
            if (_get_bit(absr, i)) {
                update_chord_aff_fp_fpxq(pdata, P, Q);
                delta = 1;
            }
        }
        i--;
    }
    pdata->do_denom_elim = old_do_denom_elim;
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
boxalletal_naf_miller_aff_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Miller variant computing f_{r,P}(Q) from the paper "A variant of
    // Miller's formula and algorithm by Boxall, El Mrabet, Laguillaumie
    // and Le, and using a non-adjacent form for r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    unsigned short old_do_denom_elim = pdata->do_denom_elim;
    pdata->do_denom_elim = 1;
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);

    int l = expi(absr) + 1;
    while ((naf[l] == 0) && (l > 0)) {
        l--;
    }
    int h = 0;
    for (int j = 0; j <= l; j ++) {
        if (naf[j] != 0) {
            h++;
        }
    }
    //pariprintf("\nNAF = [");
    //for (int j = expi(absr) + 1; j >= 0; j --) {
    //    pariprintf(" %.2i", naf[j]);
    //}
    //pariprintf("]\n\n");
    
    int i = l - 1;
    int delta = 0;
    pdata->tmp2 = FPXQ_INV(eval_vert_aff_fp_fpxq(pdata, P, Q), T, p);
    //pariprintf("\ntmp2 = %s\n", GENtostr(pdata->tmp2));
    
    if (IS_EVEN(l + h)) {
        delta = 1;
        pdata->denom = gcopy(pdata->tmp2);
    } else {
        delta = 0;
    }
    pdata->tmp = FPXQ_MUL(pdata->tmp2, ZX_Z_ADD(gel(Q, 2), gel(P, 2)), T, p);
    
    while (i >= 0) {
        if (delta == 0) {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_tangent_aff_fp_fpxq(pdata, Q);
            delta = 1;
            if (naf[i] == 1) {
                update_boxalletal_inv_if1_mm_chord_aff_fp_fpxq(pdata, P, Q);
                delta = 0;
            }
            if (naf[i] == -1) {
                update_boxalletal_inv_mp_chord_aff_fp_fpxq(pdata, P, Q);
                delta = 0;
            }
            
        } else {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_boxalletal_inv_tangent_aff_fp_fpxq(pdata, Q);
            delta = 0;
            if (naf[i] == 1) {
                update_chord_aff_fp_fpxq(pdata, P, Q);
                delta = 1;
            }
            if (naf[i] == -1) {
                update_boxalletal_if1_pm_chord_aff_fp_fpxq(pdata, P, Q);
                delta = 1;
            }
        }
        i--;
    }
    pdata->do_denom_elim = old_do_denom_elim;
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED  GEN
std_miller_aff_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Standard Miller algorithm computing f_{r,Q}(P).
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));

    GEN p = pdata->charac;
    GEN T = pdata->irredpol;

    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    long i = expi(absr) - 1;
    
    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_aff_fpxq_fp(pdata, P);

        if (_get_bit(absr, i)) {
            update_chord_aff_fpxq_fp(pdata, Q, P);
        }
        i--;
    }

    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        pdata->num = FPXQ_MUL(pdata->num, vrp, T, p);
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
        OUT();
        if (pdata->do_denom_elim) {
            return FPXQ_INV(pdata->num, T, p);
        }
        return FPXQ_DIV(pdata->denom, pdata->num, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
naf_miller_aff_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Miller algorithm computing f_{r,Q}(P) using a NAF representation of r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);

    int  i = expi(absr) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    pdata->tmp2 = FPXQ_INV(eval_vert_aff_fpxq_fp(pdata, Q, P), T, p);
    pdata->tmp  = FPXQ_MUL(pdata->tmp2, ZX_Z_ADD(gel(Q, 2), gel(P, 2)), T, p);
    
    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_aff_fpxq_fp(pdata, P);

        if (naf[i] == 1) {
            update_chord_aff_fpxq_fp(pdata, Q, P);
        } else {
            if (naf[i] == -1) {
                //
                // f_{i-1, P} = f_{i, P} * l_{iP, -P} / (v_{P} * v_{(i-1)P})
                //
                update_chord_naf_aff_fpxq_fp(pdata, Q, P);
            }
        }
        i--;
    }
    free(naf);
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        GEN tmp = pdata->denom;
     
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
        
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();    
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);

}
//----------------------------------------------------------------------------
GEN
boxalletal_miller_aff_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Miller variant computing f_{r,P}(Q) from the paper "A variant of
    // Miller's formula and algorithm by Boxall, El Mrabet, Laguillaumie
    // and Le.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    unsigned short old_do_denom_elim = pdata->do_denom_elim;
    pdata->do_denom_elim = 1;
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    
    long l = expi(absr) + 1;
    long h = 0;
    for (long j = 0; j < l; j ++) {
        h += _get_bit(absr, j);
    }
    int delta = 0;
    pdata->tmp2  = FPXQ_INV(eval_vert_aff_fpxq_fp(pdata, Q, P), T, p);
    
    if (IS_ODD(l + h)) {
        delta = 1;
        pdata->denom = gcopy(pdata->tmp2);
    } else {
        delta = 0;
    }
    pdata->tmp = FPXQ_MUL(pdata->tmp2, ZX_Z_ADD(gel(Q, 2), gel(P, 2)), T, p);
    
    long i = l - 2;
    while (i >= 0) {
        if (delta == 0) {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_tangent_aff_fpxq_fp(pdata, P);
            delta = 1;
            if (_get_bit(absr, i)) {
                update_boxalletal_inv_if1_mm_chord_aff_fpxq_fp(pdata, Q, P);
                delta = 0;
            }
            
        } else {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_boxalletal_inv_tangent_aff_fpxq_fp(pdata, P);
            delta = 0;
            if (_get_bit(absr, i)) {
                update_chord_aff_fpxq_fp(pdata, Q, P);
                delta = 1;
            }
        }
        i--;
    }
    pdata->do_denom_elim = old_do_denom_elim;
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        GEN tmp = pdata->denom;     
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
boxalletal_naf_miller_aff_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Miller variant computing f_{r,Q}(P) from the paper "A variant of
    // Miller's formula and algorithm by Boxall, El Mrabet, Laguillaumie
    // and Le, and using a non-adjacent form for r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    unsigned short old_do_denom_elim = pdata->do_denom_elim;
    pdata->do_denom_elim = 1;
     
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);

    int l = expi(absr) + 1;
    while ((naf[l] == 0) && (l > 0)) {
        l--;
    }
    int h = 0;
    for (int j = 0; j <= l; j ++) {
        if (naf[j] != 0) {
            h++;
        }
    }
    //pariprintf("\nNAF = [");
    //for (int j = expi(absr) + 1; j >= 0; j --) {
    //    pariprintf(" %.2i", naf[j]);
    //}
    //pariprintf("]\n\n");
    
    int i = l - 1;
    int delta = 0;
    pdata->tmp2 = FPXQ_INV(eval_vert_aff_fpxq_fp(pdata, Q, P), T, p);
    //pariprintf("\ntmp2 = %s\n", GENtostr(pdata->tmp2));
    
    if (IS_EVEN(l + h)) {
        delta = 1;
        pdata->denom = gcopy(pdata->tmp2);
    } else {
        delta = 0;
    }
    pdata->tmp = FPXQ_MUL(pdata->tmp2, ZX_Z_ADD(gel(Q, 2), gel(P, 2)), T, p);

    while (i >= 0) {
        if (delta == 0) {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_tangent_aff_fpxq_fp(pdata, P);
            delta = 1;
            if (naf[i] == 1) {
                update_boxalletal_inv_if1_mm_chord_aff_fpxq_fp(pdata, Q, P);
                delta = 0;
            }
            if (naf[i] == -1) {
                update_boxalletal_inv_mp_chord_aff_fpxq_fp(pdata, Q, P);
                delta = 0;
            }
            
        } else {
            pdata->num   = FPXQ_SQR(pdata->num, T, p);
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
            update_boxalletal_inv_tangent_aff_fpxq_fp(pdata, P);
            delta = 0;
            if (naf[i] == 1) {
                update_chord_aff_fpxq_fp(pdata, Q, P);
                delta = 1;
            }
            if (naf[i] == -1) {
                update_boxalletal_if1_pm_chord_aff_fpxq_fp(pdata, Q, P);
                delta = 1;
            }
        }
        i--;
    }
    pdata->do_denom_elim = old_do_denom_elim;
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_aff_fpxq_fp(pdata, pdata->point, P);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    
}
//----------------------------------------------------------------------------
