#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "pairings.h"
#include "funcs.h"
#include "commons.h"
#include "proj.h"
#include "stats.h"
#include "arithm.h"

//----------------------------------------------------------------------------
//                      Pairing in characteristic > 3
//
// WARNING: All functions in this file assume the elliptic curve is given
//          by an equation of the form y^2 = x^3 + ax + b
//
// WARNING: Excepted when mentionned otherwise, all the functions in this
//          file are NOT memory safe!
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
_get_opposite_point_proj_fp(GEN Q, GEN p) {
    //
    // Returns the opposite of the point Q on E(F_p).
    //
    // Coordinates of Q and resulting point are given by t_INTs.
    // p is a t_INT.
    //
    return mkvecn(3, gel(Q, 1), FP_NEG(gel(Q, 2), p), gel(Q, 3));
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
_get_opposite_point_proj_fpxq(GEN Q, GEN p) {
    //
    // Returns the opposite of the point Q on E(F_{p^k}).
    //
    // Coordinates of Q and resulting point are given by ZXs.
    // p is a t_INT.
    //
    return mkvecn(3, gel(Q, 1), FPX_NEG(gel(Q, 2), p), gel(Q, 3));
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE int
_are_opposite_points_proj_fp(MAYBE_UNUSED GEN E, GEN A, GEN B, GEN p) {
    IN();
    //
    // Return true if the points A and B on E(F_p) are opposites.
    // The coordinates of A and B are given by t_INTs.
    // p is a t_INT.
    //
    // WARNING -- Supposes the curve is given by y^2 = x^3 + a.x + b
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);

    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    GEN x1 = gel(A, 1);
    GEN y1 = gel(A, 2);
    GEN z1 = gel(A, 3);
    
    GEN x2 = gel(B, 1);
    GEN y2 = gel(B, 2);
    GEN z2 = gel(B, 3);
    
    GEN x1z2 = FP_MUL(x1, z2, p);
    GEN x2z1 = FP_MUL(x2, z1, p);
    GEN y1z2 = FP_MUL(y1, z2, p);
    GEN y2z1 = FP_MUL(y2, z1, p);
    
    if (!equalii(x1z2, x2z1)) {
        OUT();
        return 0;
    }
    if (equalii(y1z2, FP_NEG(y2z1, p))) {
        OUT();
        return 1;
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE int
_are_opposite_points_proj_fpxq(MAYBE_UNUSED GEN E, GEN A, GEN B, GEN t, GEN p)
{
    IN();
    //
    // Return true if the points A and B on E(F_{p^k}) are opposites.
    // The coordinates of A and B are given by FpXQs (i.e. Z_Xs).
    // t is a FpXQ (i.e. Z_X).
    // p is a t_INT.
    //
    // WARNING -- Supposes the curve is given by y^2 = x^3 + a.x + b
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);

    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    GEN x1 = gel(A, 1);
    GEN y1 = gel(A, 2);
    GEN z1 = gel(A, 3);
    
    GEN x2 = gel(B, 1);
    GEN y2 = gel(B, 2);
    GEN z2 = gel(B, 3);
    
    GEN x1z2 = FPXQ_MUL(x1, z2, t, p);
    GEN x2z1 = FPXQ_MUL(x2, z1, t, p);
    GEN y1z2 = FPXQ_MUL(y1, z2, t, p);
    GEN y2z1 = FPXQ_MUL(y2, z1, t, p);
    
    if (!ZX_equal(x1z2, x2z1)) {
        OUT();
        return 0;
    }
    if (ZX_equal(y1z2, FPX_NEG(y2z1, p))) {
        OUT();
        return 1;
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
GEN mul_proj_fp(pairing_data_t* pdata, GEN a, GEN P) {
    int* naf = _naf(a);
    int  i = expi(a) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    GEN AP = gcopy(P);
    GEN MP = _get_opposite_point_proj_fp(P, pdata->charac);
    
    while (i >= 0) {
        AP = double_proj_fp(pdata, AP);
        switch (naf[i]) {
            case 1:
                AP = add_proj_fp(pdata, AP, P);
                break;
            case -1:
                AP = add_proj_fp(pdata, AP, MP);
                break;
            default:
                break;
        }
        i--;
    }
    free(naf);
    return AP;
}
//----------------------------------------------------------------------------
GEN add_proj_fp(pairing_data_t* pdata, GEN P1, GEN P2) {
    //
    // Given P1 = (x1, y1, z1) and P2 = (x2, y2, z2) then P1 + P2 = (x3, y3, z3)
    // is given by:
    //
    //   x3 = v*A
    //   y3 = u*(R-A) - vvv*y1z2
    //   z3 = vvv*z1z2
    //
    // with:
    //
    //   y1z2 = y1*z2
    //   x1z2 = x1*z2
    //   z1z2 = z1*z2
    //   u = y2*z1 - y1z2
    //   uu = u^2
    //   v = x2*z1 - x1z2
    //   vv = v^2
    //   vvv = v*vv
    //   R = vv*x1z2
    //   A = uu*z1z2 - vvv - 2*R
    //
    // (This is the "add-1998-cmo-2" addition formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    if (ell_is_inf(P1)) {
        return gcopy(P2);
    }
    if (ell_is_inf(P2)) {
        return gcopy(P1);
    }
    GEN x1 = gel(P1, 1);
    GEN y1 = gel(P1, 2);
    GEN z1 = gel(P1, 3);
    GEN x2 = gel(P2, 1);
    GEN y2 = gel(P2, 2);
    GEN z2 = gel(P2, 3);
    
    GEN p = pdata->charac;
    
    GEN x2z1 = REM(MULII(x2, z1), p);
    GEN y2z1 = REM(MULII(y2, z1), p);

    GEN x1z2 = REM(MULII(x1, z2), p);
    GEN y1z2 = REM(MULII(y1, z2), p);
    
    if (equalii(x2z1, x1z2)) {
        if (equalii(y2z1, y1z2)) {
            return double_proj_fp(pdata, P1);
        } else {
            return ellinf();
        }
    }
    GEN z1z2 = MULII(z1, z2);
    GEN u    = SUBII(y2z1, y1z2);
    GEN uu   = FP_SQR(u, p);
    GEN v    = SUBII(x2z1, x1z2);
    GEN vv   = FP_SQR(v, p);
    GEN vvv  = MULII(v, vv);
    GEN R    = MULII(vv, x1z2);
    GEN A    = SUBII(SUBII(MULII(uu, z1z2), vvv), MULUI(2, R));
        
    GEN x3 = FP_MUL(v, A, p);
    GEN y3 = FP_SUB(MULII(u, SUBII(R, A)), MULII(vvv, y1z2), p);
    GEN z3 = FP_MUL(vvv, z1z2, p);
    
    x3 = Rg_to_Fp(x3, p);
    y3 = Rg_to_Fp(y3, p);
    z3 = Rg_to_Fp(z3, p);
    
    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN double_proj_fp(pairing_data_t* pdata, GEN P) {
    //
    // Given P = (x1, y1, z1) then [2]P = (x3, y3, z3) is given by:
    //
    //   x3 = h*s
    //   y3 = w*(b-h)-2*rr
    //   z3 = sss
    //
    // with:
    //
    //   xx  = x1^2
    //   zz  = z1^2
    //   w   = a * zz + 3 * xx
    //   s   = 2 * y1 * z1
    //   ss  = s^2
    //   sss = s * ss
    //   r   = y1*s
    //   rr  = r^2
    //   b   = (x1 + r)^2 - xx - rr
    //   h   = w^2 - 2 * b
    //
    if (ell_is_inf(P)) {    
        return ellinf();
    }
    GEN x1 = gel(P, 1);
    GEN y1 = gel(P, 2);
    GEN z1 = gel(P, 3);
    
    if (cmpis(y1, 0) == 0) {
        return ellinf();
    }
    
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN a = gel(gel(E, 4), 2);
    
    GEN xx = SQRI(x1);
    GEN zz = SQRI(z1);
    
    GEN w   = ADDII(MULII(a, zz), MULUI(3, xx));
    GEN s   = MULUI(2, MULII(y1, z1));
    GEN ss  = SQRI(s);
    GEN sss = MULII(s, ss);
    GEN r   = MULII(y1, s);
    GEN rr  = SQRI(r);
    GEN b   = SUBII(SUBII(SQRI(ADDII(x1, r)), xx), rr);
    GEN h   = SUBII(SQRI(w), MULUI(2, b));

    GEN x3 = MULII(h, s);
    GEN y3 = SUBII(MULII(w, SUBII(b, h)), MULUI(2, rr));
    GEN z3 = sss;
    
    x3 = Rg_to_Fp(x3, p);
    y3 = Rg_to_Fp(y3, p);
    z3 = Rg_to_Fp(z3, p);
    
    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN mul_proj_fpxq(pairing_data_t* pdata, GEN a, GEN Q) {
    int* naf = _naf(a);
    int  i = expi(a) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    GEN AQ = gcopy(Q);
    GEN MQ = _get_opposite_point_proj_fpxq(Q, pdata->charac);
    
    while (i >= 0) {
        AQ = double_proj_fpxq(pdata, AQ);
        switch (naf[i]) {
            case 1:
                AQ = add_proj_fpxq(pdata, AQ, Q);
                break;
            case -1:
                AQ = add_proj_fpxq(pdata, AQ, MQ);
                break;
            default:
                break;
        }
        i--;
    }
    free(naf);
    return AQ;
}
//----------------------------------------------------------------------------
GEN add_proj_fpxq(pairing_data_t* pdata, GEN Q1, GEN Q2) {
    //
    // Given Q1 = (x1, y1, z1) and Q2 = (x2, y2, z2) then Q1 + Q2 = (x3, y3, z3)
    // is given by:
    //
    //   x3 = v*A
    //   y3 = u*(R-A) - vvv*y1z2
    //   z3 = vvv*z1z2
    //
    // with:
    //
    //   y1z2 = y1*z2
    //   x1z2 = x1*z2
    //   z1z2 = z1*z2
    //   u = y2*z1 - y1z2
    //   uu = u^2
    //   v = x2*z1 - x1z2
    //   vv = v^2
    //   vvv = v*vv
    //   R = vv*x1z2
    //   A = uu*z1z2 - vvv - 2*R
    //
    // (This is the "add-1998-cmo-2" addition formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    if (ell_is_inf(Q1)) {
        return gcopy(Q2);
    }
    if (ell_is_inf(Q2)) {
        return gcopy(Q1);
    }
    GEN x1 = gel(Q1, 1);
    GEN y1 = gel(Q1, 2);
    GEN z1 = gel(Q1, 3);
    GEN x2 = gel(Q2, 1);
    GEN y2 = gel(Q2, 2);
    GEN z2 = gel(Q2, 3);
    
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x2z1 = FPXQ_MUL(x2, z1, t, p);
    GEN y2z1 = FPXQ_MUL(y2, z1, t, p);
    GEN x1z2 = FPXQ_MUL(x1, z2, t, p);
    GEN y1z2 = FPXQ_MUL(y1, z2, t, p);
    
    if (_ZX_equal(x2z1, x1z2)) {
        if (_ZX_equal(y2z1, y1z2)) {
            return double_proj_fpxq(pdata, Q1);
        } else {
            return ellinf();
        }
    }
    GEN z1z2 = FPXQ_MUL(z1, z2, t, p);
    GEN u    = ZX_SUB(y2z1, y1z2);
    GEN uu   = FPXQ_SQR(u, t, p);
    GEN v    = ZX_SUB(x2z1, x1z2);
    GEN vv   = FPXQ_SQR(v, t, p);
    GEN vvv  = FPXQ_MUL(v, vv, t, p);
    GEN R    = FPXQ_MUL(vv, x1z2, t, p);
    GEN A    = ZX_SUB(ZX_SUB(FPXQ_MUL(uu, z1z2, t, p), vvv), ZX_MULS(R, 2));
    
    GEN x3 = FPXQ_MUL(v, A, t, p);
    GEN y3 = ZX_SUB(FPXQ_MUL(u, ZX_SUB(R, A), t, p), FPXQ_MUL(vvv, y1z2, t, p));
    GEN z3 = FPXQ_MUL(vvv, z1z2, t, p);
    
    x3 = Rg_to_FpXQ(x3, t, p);
    y3 = Rg_to_FpXQ(y3, t, p);
    z3 = Rg_to_FpXQ(z3, t, p);

    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN double_proj_fpxq(pairing_data_t* pdata, GEN Q) {
    //
    // Given Q = (x1, y1, z1) then [2]Q = (x3, y3, z3) is given by:
    //
    //   x3 = h*s
    //   y3 = w*(b-h)-2*rr
    //   z3 = sss
    //
    // with:
    //
    //   xx  = x1^2
    //   zz  = z1^2
    //   w   = a * zz + 3 * xx
    //   s   = 2 * y1 * z1
    //   ss  = s^2
    //   sss = s * ss
    //   r   = y1*s
    //   rr  = r^2
    //   b   = (x1 + r)^2 - xx - rr
    //   h   = w^2 - 2 * b
    //
    if (ell_is_inf(Q)) {
        return ellinf();
    } 
    GEN x1 = gel(Q, 1);
    GEN y1 = gel(Q, 2);
    GEN z1 = gel(Q, 3);
    
    if (signe(y1) == 0) {
        return ellinf();
    }
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN a = gel(gel(E, 4), 2);
    
    GEN xx = FPXQ_SQR(x1, t, p);
    GEN zz = FPXQ_SQR(z1, t, p);
    
    GEN w   = ZX_ADD(FPX_FP_MUL(zz, a, p), ZX_MULS(xx, 3));
    GEN s   = ZX_MULS(FPXQ_MUL(y1, z1, t, p), 2);
    GEN ss  = FPXQ_SQR(s, t, p);
    GEN sss = FPXQ_MUL(s, ss, t, p);
    GEN r   = FPXQ_MUL(y1, s, t, p);
    GEN rr  = FPXQ_SQR(r, t, p);
    GEN b   = ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(x1, r), t, p), xx), rr);
    GEN h   = ZX_SUB(FPXQ_SQR(w, t, p), ZX_MULS(b, 2));

    GEN x3 = FPXQ_MUL(h, s, t, p);
    GEN y3 = ZX_SUB(FPXQ_MUL(w, ZX_SUB(b, h), t, p), ZX_MULS(rr, 2));
    GEN z3 = sss;
    
    x3 = Rg_to_FpXQ(x3, t, p);
    y3 = Rg_to_FpXQ(y3, t, p);
    z3 = Rg_to_FpXQ(z3, t, p);
    
    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN
eval_vert_proj_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // A is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // Return value is a ZX.
    //
    // Supposes that Q = (X, Y, 1).
    //
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    if (ell_is_inf(A)) {
        OUT();
        return Rg_to_FpXQ(gen_1, t, p);
    }
    GEN V = FPX_FP_MUL(gel(Q, 1), gel(A, 3), p);
    V = FPX_FP_ADD(V, FP_NEG(gel(A, 1), p), p);
    
    OUT();
    return V;
}
//----------------------------------------------------------------------------
void
update_vert_proj_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // A is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // Return value is a ZX.
    //
    // Supposes that Q = (X, Y, 1).
    //
    if (pdata->do_denom_elim) {
        return;
    }
    if (!ell_is_inf(A)) {
        GEN p = pdata->charac;
        GEN t = pdata->irredpol;
        
        GEN V = FPX_FP_MUL(gel(Q, 1), gel(A, 3), p);
        V = FPX_FP_ADD(V, FP_NEG(gel(A, 1), p), p);
        
        pdata->denom = FPXQ_MUL(pdata->denom, V, t, p);
        OUT();
        return;
    }
}
//----------------------------------------------------------------------------
void
update_tangent_proj_fp_fpxq(pairing_data_t* pdata, GEN Q)
{
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point Q to update pf->num and finally doubles the point
    // pf->point.
    //
    // pf.point is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P = (x1, y1, z1) then [2]P = (x3, y3, z3) is given by:
    //
    //   x3 = h*s
    //   y3 = w*(b-h)-2*rr
    //   z3 = sss
    //
    // with:
    //
    //   xx  = x1^2
    //   zz  = z1^2
    //   w   = a * zz + 3 * xx
    //   s   = 2 * y1 * z1
    //   ss  = s^2
    //   sss = s * ss
    //   r   = y1*s
    //   rr  = r^2
    //   b   = (x1 + r)^2 - xx - rr
    //   h   = w^2 - 2 * b
    //
    // (This is the "dbl-2007-bl" doubling formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    // The line tangent to [2]P has for equation:
    //
    //     L(PP) : Y - slope * (X - x1/z1) - y1/z1  (= 0)
    //
    // with: slope = w / s
    //
    // The computed line here is taken as:
    //
    //    CL(PP) = 2.y1.z1^2.L(PP)
    //           = s.z1.Y - w.(z1.X - x1) - s.y1
    //
    // We suppose that the line is evaluated at (XQ, YQ, 1). For Tate-like 
    // pairings the s.z1 will cancel out during the final exponentation.
    //
    IN();
    if (ell_is_inf(pdata->point)) {    
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    } 
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);
    GEN z1 = gel(pdata->point, 3);
    
    if (cmpis(y1, 0) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        pdata->num = FPXQ_MUL(
                       pdata->num,
                       eval_vert_proj_fp_fpxq(pdata, pdata->point, Q),
                       t, p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    GEN a = gel(gel(E, 4), 2);
    
    //   x3 = h * s
    //   y3 = w * (b - h) - 2 * rr
    //   z3 = sss
    //
    // with:
    //
    //   xx  = x1^2
    //   zz  = z1^2
    //   w   = a * zz + 3 * xx
    //   s   = 2 * y1 * z1
    //   ss  = s^2
    //   sss = s * ss
    //   r   = y1*s
    //   rr  = r^2
    //   b   = (x1 + r)^2 - xx - rr
    //   h   = w^2 - 2 * b
    //
    GEN xx = SQRI(x1);
    GEN zz = SQRI(z1);
    
    GEN w   = ADDII(MULII(a, zz), MULUI(3, xx));
    GEN s   = MULUI(2, MULII(y1, z1));
    GEN ss  = SQRI(s);
    GEN sss = MULII(s, ss);
    GEN r   = MULII(y1, s);
    GEN rr  = SQRI(r);
    GEN b   = SUBII(SUBII(SQRI(ADDII(x1, r)), xx), rr);
    GEN h   = SUBII(SQRI(w), MULUI(2, b));

    GEN x3 = MULII(h, s);
    GEN y3 = SUBII(MULII(w, SUBII(b, h)), MULUI(2, rr));
    GEN z3 = sss;

    //
    //    CL(PP) = 2.y1.z1^2.L(PP)
    //           = s.z1.Y - w.(z1.X - x1) - s.y1
    //           = s.z1.Y - w.z1.X + w.x1 - s.y1
    //           = la - lb + lc - ld
    //    
    GEN la = FPX_FP_MUL(gel(Q, 2), MULII(s, z1), p);
    GEN lb = FPX_FP_MUL(gel(Q, 1), MULII(w, z1), p);
    GEN lc = MULII(x1, w);
    GEN ld = MULII(s, y1);

    GEN line = FPX_FP_ADD(
                 FPX_ADD(la, FPX_NEG(lb, p), p),
                 SUBII(lc, ld),
                 p
               );
    
    gel(pdata->point, 1) = Rg_to_Fp(x3, p);
    gel(pdata->point, 2) = Rg_to_Fp(y3, p);
    gel(pdata->point, 3) = Rg_to_Fp(z3, p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);
    
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                         pdata->denom,
                         eval_vert_proj_fp_fpxq(pdata, pdata->point, Q),
                         t,
                         p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_proj_fp_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Computes the equation of the line through pf.point and P (supposed to
    // be different from the point at infinity), evaluates it at the point Q
    // to update pf->num, adds P to the point pfunc->point and finally
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point Q.
    //
    // pf.point and P are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1, z1) and P2 = (x2, y2, 1) then P1 + P2 = (x3, y3, z3)
    // is given by:
    //
    //     x3 = v.A
    //     y3 = u.(R - A) - vvv.y1;
    //     z3 = vvv.z1;
    //
    // with:
    // 
    //     u   = y2.z1 - y1
    //     uu  = u^2
    //     v   = x2.z1 - x1
    //     vv  = v.v
    //     vvv = v.vv 
    //     R   = vv.x1
    //     A   = uu.z1 - vvv - 2.R
    //
    // (This is the "madd-1998-cmo" addition formula, as given by the explicit 
    //  formula database at http://www.hyperelliptic.org/EFD)
    //
    // The line L(P1P2) through P1 and P2 has for equation:
    //
    //     L(P1P2) : Y - slope * (X - x2) - y2  (= 0) (remember z2 = 1)
    //
    // with: slope = u / v
    //
    // The computed line here is taken as:
    //
    //    CL(P1P2) = v.L(P1P2)
    //             = v.Y - u.(X - x2) - v.y2
    //
    // We suppose that the line is evaluated at (XQ, YQ, 1). For Tate-like 
    // pairings the v will cancel out during the final exponentation.
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;

    int t_is_inf = ell_is_inf(T);
    if (t_is_inf) {
        //
        // The chord is the vertical line passing through T
        //
        pdata->num = FPXQ_MUL(
                         pdata->num,
                         eval_vert_proj_fp_fpxq(pdata, T, Q),
                         t,
                         p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    //
    // From now on we know that both T and P are not the point at inifinity.
    //
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);
    GEN z1 = gel(T, 3);
    
    GEN x2 = gel(P, 1);
    GEN y2 = gel(P, 2);

    GEN y2z1 = REM(MULII(y2, z1), p);
    GEN x2z1 = REM(MULII(x2, z1), p);
    
    if (equalii(x2z1, x1)) {
        //
        // T and P are on the same vertical line.
        //      
        if (equalii(y2z1, y1)) {
            //
            // T and P are the same point and are different from the point
            // at inifinity.
            //
            update_tangent_proj_fp_fpxq(pdata, Q);
            OUT();
            return;
            
        } else {
            //
            // The chord is the vertical line passing through T
            //
            pdata->num = FPXQ_MUL(
                             pdata->num,
                             eval_vert_proj_fp_fpxq(pdata, T, Q),
                             t,
                             p
                         );
            pdata->point = ellinf();
            OUT();
            return;
        }
    }
    //
    //     x3 = v.A
    //     y3 = u.(R - A) - vvv.y1;
    //     z3 = vvv.z1;
    //
    // with:
    // 
    //     u   = y2.z1 - y1
    //     uu  = u^2
    //     v   = x2.z1 - x1
    //     vv  = v.v
    //     vvv = v.vv 
    //     R   = vv.x1
    //     A   = uu.z1 - vvv - 2.R
    //
    GEN u    = SUBII(y2z1, y1);
    GEN uu   = FP_SQR(u, p);
    GEN v    = SUBII(x2z1, x1);
    GEN vv   = FP_SQR(v, p);
    GEN vvv  = MULII(v, vv);
    GEN R    = MULII(vv, x1);
    GEN A    = SUBII(SUBII(MULII(uu, z1), vvv), MULUI(2, R));
    
    GEN x3 = FP_MUL(v, A, p);
    GEN y3 = FP_SUB(MULII(u, SUBII(R, A)), MULII(vvv, y1), p);
    GEN z3 = FP_MUL(vvv, z1, p);
    
    //
    //    CL(P1P2) = v.L(P1P2)
    //             = v.(Y - y2) - u.(X - x2)
    //             = la - lb
    //
    GEN la = FPX_FP_ADD(gel(Q, 2), FP_NEG(y2, p), p);
        la = FPX_FP_MUL(la, v, p);

    GEN lb = FPX_FP_ADD(gel(Q, 1), FP_NEG(x2, p), p);
        lb = FPX_FP_MUL(lb, u, p);
        
    GEN line = FPX_ADD(la, FPX_NEG(lb, p), p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = Rg_to_Fp(x3, p);
    gel(pdata->point, 2) = Rg_to_Fp(y3, p);
    gel(pdata->point, 3) = Rg_to_Fp(z3, p);
        
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                           pdata->denom,
                           eval_vert_proj_fp_fpxq(pdata, pdata->point, Q),
                           t,
                           p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
GEN
eval_vert_proj_fpxq_fp(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // pf.point and A are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. Q is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Supposes that Q = (X, Y, 1).
    //
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    if (ell_is_inf(A)) {
        OUT();
        return Rg_to_FpXQ(gen_1, t, p);
    }
    GEN V = FPX_FP_MUL(gel(A, 3), gel(Q, 1), p);
    V = FPX_ADD(V, FPX_NEG(gel(A, 1), p), p);
    
    OUT();
    return V;
}
//----------------------------------------------------------------------------
void
update_vert_proj_fpxq_fp(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // pf.point and A are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. Q is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Supposes that Q = (X, Y, 1).
    //
    if (pdata->do_denom_elim) {
        return;
    }
    if (!ell_is_inf(A)) {
        GEN p = pdata->charac;
        GEN t = pdata->irredpol;
        
        GEN V = FPX_FP_MUL(gel(A, 3), gel(Q, 1), p);
        V = FPX_ADD(V, FPX_NEG(gel(A, 1), p), p);
        
        pdata->denom = FPXQ_MUL(pdata->denom, V, t, p);
        OUT();
        return;
    }
}
//----------------------------------------------------------------------------
void
update_tangent_proj_fpxq_fp(pairing_data_t* pdata, GEN P)
{
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point P to update pf->num and finally doubles the point
    // pf->point.
    //
    // pf.point is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given Q = (x1, y1, z1) then [2]Q = (x3, y3, z3) is given by:
    //
    //   x3 = h*s
    //   y3 = w*(b-h)-2*rr
    //   z3 = sss
    //
    // with:
    //
    //   xx  = x1^2
    //   zz  = z1^2
    //   w   = a * zz + 3 * xx
    //   s   = 2 * y1 * z1
    //   ss  = s^2
    //   sss = s * ss
    //   r   = y1*s
    //   rr  = r^2
    //   b   = (x1 + r)^2 - xx - rr
    //   h   = w^2 - 2 * b
    //
    // (This is the "dbl-2007-bl" doubling formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    // The line tangent to [2]P has for equation:
    //
    //     L(PP) : Y - slope * (X - x1/z1) - y1/z1  (= 0)
    //
    // with: slope = w / s
    //
    // The computed line here is taken as:
    //
    //    CL(PP) = 2.y1.z1^2.L(PP)
    //           = s.z1.Y - w.(z1.X - x1) - s.y1
    //
    // We suppose that the line is evaluated at (XP, YP, 1). For Tate-like 
    // pairings the 2.y1.z1^2 will cancel out during the final exponentation.
    //
    IN();
    if (ell_is_inf(pdata->point)) {
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    } 
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);
    GEN z1 = gel(pdata->point, 3);
    
    if (signe(y1) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        pdata->num = FPXQ_MUL(
                       pdata->num,
                       eval_vert_proj_fpxq_fp(pdata, pdata->point, P),
                       t, p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    GEN a = gel(gel(E, 4), 2);
    
    //   x3 = h * s
    //   y3 = w * (b - h) - 2 * rr
    //   z3 = sss
    //
    // with:
    //
    //   xx  = x1^2
    //   zz  = z1^2
    //   w   = a * zz + 3 * xx
    //   s   = 2 * y1 * z1
    //   ss  = s^2
    //   sss = s * ss
    //   r   = y1*s
    //   rr  = r^2
    //   b   = (x1 + r)^2 - xx - rr
    //   h   = w^2 - 2 * b
    //
    GEN xx = FPXQ_SQR(x1, t, p);
    GEN zz = FPXQ_SQR(z1, t, p);
    
    GEN w   = ZX_ADD(FPX_FP_MUL(zz, a, p), ZX_MULS(xx, 3));
    GEN s   = ZX_MULS(FPXQ_MUL(y1, z1, t, p), 2);
    GEN ss  = FPXQ_SQR(s, t, p);
    GEN sss = FPXQ_MUL(s, ss, t, p);
    GEN r   = FPXQ_MUL(y1, s, t, p);
    GEN rr  = FPXQ_SQR(r, t, p);
    GEN b   = ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(x1, r), t, p), xx), rr);
    GEN h   = ZX_SUB(FPXQ_SQR(w, t, p), ZX_MULS(b, 2));

    GEN x3 = FPXQ_MUL(h, s, t, p);
    GEN y3 = ZX_SUB(FPXQ_MUL(w, ZX_SUB(b, h), t, p), ZX_MULS(rr, 2));
    GEN z3 = sss;

    //
    //    CL(PP) = 2.y1.z1^2.L(PP)
    //           = s.z1.Y - w.(z1.X - x1) - s.y1
    //           = s.z1.Y - w.z1.X + w.x1 - s.y1
    //           = la - lb + lc - ld
    //    
    GEN la = FPX_FP_MUL(FPXQ_MUL(s, z1, t, p), gel(P, 2), p);    
    GEN lb = FPX_FP_MUL(FPXQ_MUL(w, z1, t, p), gel(P, 1), p);
    GEN lc = FPXQ_MUL(x1, w, t, p);
    GEN ld = FPXQ_MUL(s, y1, t, p);

    GEN line = FPX_ADD(
                 FPX_ADD(la, FPX_NEG(lb, p), p),
                 FPX_ADD(lc, FPX_NEG(ld, p), p),
                 p
               );
    
    gel(pdata->point, 1) = Rg_to_FpXQ(x3, t, p);
    gel(pdata->point, 2) = Rg_to_FpXQ(y3, t, p);
    gel(pdata->point, 3) = Rg_to_FpXQ(z3, t, p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);
    
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                         pdata->denom,
                         eval_vert_proj_fpxq_fp(pdata, pdata->point, P),
                         t,
                         p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_proj_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Computes the equation of the line through pf.point and Q (supposed to
    // be different from the point at infinity), evaluates it at the point P
    // to update pf->num, adds Q to the point pfunc->point and finally
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point P.
    //
    // pf.point and Q are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given P1 = (x1, y1, z1) and P2 = (x2, y2, 1) then P1 + P2 = (x3, y3, z3)
    // is given by:
    //
    //     x3 = v.A
    //     y3 = u.(R - A) - vvv.y1;
    //     z3 = vvv.z1;
    //
    // with:
    // 
    //     u   = y2.z1 - y1
    //     uu  = u^2
    //     v   = x2.z1 - x1
    //     vv  = v.v
    //     vvv = v.vv 
    //     R   = vv.x1
    //     A   = uu.z1 - vvv - 2.R
    //
    // (This is the "madd-1998-cmo" addition formula, as given by the explicit 
    //  formula database at http://www.hyperelliptic.org/EFD)
    //
    // The line L(P1P2) through P1 and P2 has for equation:
    //
    //     L(P1P2) : Y - slope * (X - x2) - y2  (= 0) (remember z2 = 1)
    //
    // with: slope = u / v
    //
    // The computed line here is taken as:
    //
    //    CL(P1P2) = v.L(P1P2)
    //             = v.Y - u.(X - x2) - v.y2
    //
    // We suppose that the line is evaluated at (XP, YP, 1). For Tate-like 
    // pairings the v will cancel out during the final exponentation.
    //
    IN();

    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;

    int t_is_inf = ell_is_inf(T);
    if (t_is_inf) {
        //
        // The chord is the vertical line passing through T
        //
        pdata->num = FPXQ_MUL(
                         pdata->num,
                         eval_vert_proj_fpxq_fp(pdata, T, P),
                         t,
                         p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    //
    // From now on we know that both T and P are not the point at inifinity.
    //
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);
    GEN z1 = gel(T, 3);
    
    GEN x2 = gel(Q, 1);
    GEN y2 = gel(Q, 2);

    GEN y2z1 = FPXQ_MUL(y2, z1, t, p);
    GEN x2z1 = FPXQ_MUL(x2, z1, t, p);
    
    if (_ZX_equal(x2z1, x1) == 0) {
        //
        // T and P are on the same vertical line.
        //      
        if (_ZX_equal(y2z1, y1)) {
            //
            // T and Q are the same point and are different from the point
            // at inifinity.
            //
            update_tangent_proj_fpxq_fp(pdata, P);
            OUT();
            return;
            
        } else {
            //
            // The chord is the vertical line passing through T
            //
            pdata->num = FPXQ_MUL(
                             pdata->num,
                             eval_vert_proj_fpxq_fp(pdata, T, P),
                             t,
                             p
                         );
            pdata->point = ellinf();
            OUT();
            return;
        }
    }
    //
    //     x3 = v.A
    //     y3 = u.(R - A) - vvv.y1;
    //     z3 = vvv.z1;
    //
    // with:
    // 
    //     u   = y2.z1 - y1
    //     uu  = u^2
    //     v   = x2.z1 - x1
    //     vv  = v.v
    //     vvv = v.vv 
    //     R   = vv.x1
    //     A   = uu.z1 - vvv - 2.R
    //
    GEN u    = ZX_SUB(y2z1, y1);
    GEN uu   = FPXQ_SQR(u, t, p);
    GEN v    = ZX_SUB(x2z1, x1);
    GEN vv   = FPXQ_SQR(v, t, p);
    GEN vvv  = FPXQ_MUL(v, vv, t, p);
    GEN R    = FPXQ_MUL(vv, x1, t, p);
    GEN A    = ZX_SUB(ZX_SUB(FPXQ_MUL(uu, z1, t, p), vvv), ZX_MULS(R, 2));
    
    GEN x3 = FPXQ_MUL(v, A, t, p);
    GEN y3 = ZX_SUB(FPXQ_MUL(u, ZX_SUB(R, A), t, p), FPXQ_MUL(vvv, y1, t, p));
    GEN z3 = FPXQ_MUL(vvv, z1, t, p);
    
    //
    //    CL(P1P2) = v.L(P1P2)
    //             = v.(Y - y2) - u.(X - x2)
    //             = la - lb
    //
    GEN la = FPX_FP_ADD(FPX_NEG(y2, p), gel(P, 2), p);
        la = FPXQ_MUL(la, v, t, p);

    GEN lb = FPX_FP_ADD(FPX_NEG(x2, p), gel(P, 1), p);
        lb = FPXQ_MUL(lb, u, t, p);
        
    GEN line = FPX_SUB(la, lb, p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = Rg_to_FpXQ(x3, t, p);
    gel(pdata->point, 2) = Rg_to_FpXQ(y3, t, p);
    gel(pdata->point, 3) = Rg_to_FpXQ(z3, t, p);
        
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                           pdata->denom,
                           eval_vert_proj_fpxq_fp(pdata, pdata->point, P),
                           t,
                           p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
MAYBE_UNUSED GEN
std_miller_proj_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Standard Miller algorithm computing f_{r,P}(Q).
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    long i = expi(absr) - 1;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_proj_fp_fpxq(pdata, Q);

        if (_get_bit(absr, i)) {
            update_chord_proj_fp_fpxq(pdata, P, Q);
        }
        i--;
    }
    OUT();
    if (r_is_neg) {
        GEN vrp = eval_vert_proj_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    
    if (pdata->do_denom_elim) {
        return pdata->num;
    }
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
naf_miller_proj_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Miller algorithm computing f_{r,P}(Q) using a NAF representation of r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));

    GEN p = pdata->charac;
    GEN T = pdata->irredpol;

    GEN minusP = _get_opposite_point_proj_fp(P, p);
    
    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);

    int  i = expi(absr) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_proj_fp_fpxq(pdata, Q);

        if (naf[i] == 1) {
            update_chord_proj_fp_fpxq(pdata, P, Q);
        } else {
            if (naf[i] == -1) {
                //
                // f_{i-1, P} = f_{i, P} * l_{iP, -P} / (v_{1, P} * v_{(i-1)P})
                //
                update_chord_proj_fp_fpxq(pdata, minusP, Q);
                
                if (!pdata->do_denom_elim) {
                    pdata->denom = FPXQ_MUL(
                                     pdata->denom,
                                     eval_vert_proj_fp_fpxq(pdata, P, Q),
                                     T,
                                     p
                                   );
                }
            }
        }
        i--;
    }
    free(naf);
    if (r_is_neg) {
        GEN vrp = eval_vert_proj_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED  GEN
std_miller_proj_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Standard Miller algorithm computing f_{r,Q}(P).
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));

    GEN p = pdata->charac;
    GEN T = pdata->irredpol;

    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    long i = expi(absr) - 1;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_proj_fpxq_fp(pdata, P);

        if (_get_bit(absr, i)) {
            update_chord_proj_fpxq_fp(pdata, Q, P);
        }
        i--;
    }
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_proj_fpxq_fp(pdata, pdata->point, P);
        pdata->num = FPXQ_MUL(pdata->num, vrp, T, p);
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
        OUT();
        if (pdata->do_denom_elim) {
            return FPXQ_INV(pdata->num, T, p);
        }
        return FPXQ_DIV(pdata->denom, pdata->num, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
naf_miller_proj_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Miller algorithm computing f_{r,Q}(P) using a NAF representation of r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    GEN minusQ = _get_opposite_point_proj_fpxq(Q, p);

    int r_is_neg = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);

    int  i = expi(absr) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_proj_fpxq_fp(pdata, P);

        if (naf[i] == 1) {
            update_chord_proj_fpxq_fp(pdata, Q, P);
        } else {
            if (naf[i] == -1) {
                //
                // f_{i-1, P} = f_{i, P} * l_{iP, -P} / (v_{1, P} * v_{(i-1)P})
                //
                update_chord_proj_fpxq_fp(pdata, minusQ, P);
                pdata->denom = FPXQ_MUL(
                                 pdata->denom,
                                 eval_vert_proj_fpxq_fp(pdata, Q, P),
                                 T,
                                 p
                               );
            }
        }
        i--;
    }
    free(naf);
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_proj_fpxq_fp(pdata, pdata->point, P);
        GEN tmp = pdata->denom;
        
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
