#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "pairings.h"
#include "funcs.h"

//----------------------------------------------------------------------------
// WARNING: Excepted when mentionned otherwise, all the functions in this
//          file are NOT memory safe!
//----------------------------------------------------------------------------
GEN _ZX_muls(GEN y, long x) {
    //
    // Multiply y (given as a ZX) by the long integer x and returns the
    // result as a new ZX.
    //
    GEN z;
    long i, l;
    if (x == 0) {
        return zeropol(varn(y));
    }
    l = lg(y);
    z = cgetg(l, t_POL);
    z[1] = y[1];
    
    for (i = 2; i < l; i++) {
        gel(z, i) = mulis(gel(y, i), x);
    }
    return z;
}
//----------------------------------------------------------------------------
GEN _FpXQ_to_FF(GEN v, GEN T, GEN p) {
    //
    // Returns v, an element of F_{p^k} given as an FpXQ, as a t_FFELT
    // of format t_FF_FpXQ.
    //
    // T is the irreducible t_POL (in F_p) used to define F_{p^k}.
    // p is the characteristic of the involved fields.
    //
    GEN elt = cgetg(5, t_FFELT);
    elt[1]  = t_FF_FpXQ;
    gel(elt, 2) = gcopy(v);
    gel(elt, 3) = gcopy(T);
    gel(elt, 4) = icopy(p);
    setvarn(gel(elt, 3), varn(v));
    return elt;
}
//----------------------------------------------------------------------------
GEN _Flxq_to_FF(GEN v, GEN T, long l) {
    //
    // Returns v, an element of F_{p^k} given as an Flxq, as a t_FFELT
    // of format t_FF_Flxq.
    //
    // T is the irreducible t_POL (in F_p) used to define F_{p^k}.
    // p is th characteristic of the involved fields.
    //
    GEN elt = cgetg(5, t_FFELT);
    elt[1]  = t_FF_Flxq;
    gel(elt, 2) = gcopy(v);
    gel(elt, 3) = gcopy(T);
    gel(elt, 4) = mkintn(1, l);
    return elt;
}
//----------------------------------------------------------------------------
GEN _F2xq_to_FF(GEN v, GEN T) {
    //
    // Returns v, an element of F_{2^k} given as an F2xq, as a t_FFELT
    // of format t_FF_F2xq.
    //
    // T is the irreducible t_POL (in F_2) used to define F_{2^k}.
    //
    GEN elt = cgetg(5, t_FFELT);
    elt[1]  = t_FF_F2xq;
    gel(elt, 2) = gcopy(v);
    gel(elt, 3) = gcopy(T);
    gel(elt, 4) = mkintn(1, 2);
    return elt;
}
//----------------------------------------------------------------------------
int _are_same_point(GEN A, GEN B) {
    //
    // Return true if the points A and B (given in affine coordinate) are
    // the same ie returns 1 iif (x_a == x_b) && (y_a == y_b), 0 otherwise.
    // This is a generic function. A and B must have component of GEN type.
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);
    
    if (a_is_inf && b_is_inf) {
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        return 0;
    }
    if (gequal(gel(A, 1), gel(B, 1)) && gequal(gel(A, 2), gel(B, 2))) {
        return 1;
    }
    return 0;
}
//----------------------------------------------------------------------------
int _are_opposite_points(GEN E, GEN A, GEN B) {
    IN();
    //
    // Return true if the points A and B (given in affine coordinate) are 
    // opposites ie returns 1 iif A = (x, y) and B = (x, -y - a1.x - a3).
    // This is a generic function. A and B must have component of GEN type.
    //    
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);
    
    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    if (!gequal(gel(A, 1), gel(B, 1))) {
        OUT();
        return 0;
    }
    if (gcmp0(gel(E, 1)) && gcmp0(gel(E, 3))) {
        if (gequal(gel(A, 2), gneg(gel(B, 2)))) {
            OUT();
            return 1;
        }
        OUT();
        return 0;
    } else {
        GEN yneg = gadd(gel(B, 2), gmul(gel(B, 1), gel(E,1)));
        yneg = gadd(yneg, gel(E,3));
        if (gequal(gel(A, 2), gneg(yneg))) {
            OUT();
            return 1;
        }
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
GEN _intmod_val(GEN x) {
    //
    // Returns a t_INTMODs are a newly allocated t_INT (i.e. strips the
    // modulo).
    //
    return gel(x, 2);
}
//----------------------------------------------------------------------------
ulong _get_bit(GEN x, long n) {
    //
    // Return the value of the n-th bit of the t_INT x as 0 or 1.
    //
    long r;
    long q = dvmdsBIL(n, &r);
    return (*int_W(x, q) >> r) & 1;
}
//----------------------------------------------------------------------------
GEN _get_char(GEN a) {
    //
    // Returns the characteristic of the field element a as a t_INT if and
    // only if a is given as a t_FFELT (for prime field extensions) or a 
    // t_INTMOD (for prime fields). Otherwise, returns 0 as a t_INT.
    //
    long t = typ(a);
    switch (t) {
        case t_FFELT:
            return FF_p(a);
            break;
        case t_INTMOD:
            return gel(a, 1);
            break;
        default:
            ERR("not a field element");
            return mkintn(1, 0);
    }
}
//----------------------------------------------------------------------------
int _ZX_equal(GEN x, GEN y) {
    //
    // Shamelessly stolen and adapted from the function polequal in 
    // src/basemath/gen2.c since the function there is declared static
    // for some reason... This version is restricted to t_POLs with
    // t_INT coefficients.
    //
    long lx, ly;
    if (x[1] != y[1]) {
        return 0;
    }
    lx = lg(x);
    ly = lg(y);

    while (lx > ly) {
        if (!equalis(gel(x,--lx), 0)) {
            return 0;
        }
    }
    while (ly > lx) {
        if (!equalis(gel(y,--ly), 0)) {
            return 0;
        }
    }
    for (lx--; lx >= 2; lx--) {
        if (!equalii(gel(x,lx), gel(y,lx))) {
            return 0;
        }
    }
    return 1;
}
//----------------------------------------------------------------------------
int* _naf(GEN k) {
    //
    // Computes the non adjacent form of the l-bit number k and returns it as
    // an array b of (l+1) bytes, with b[i] = -1, 0 or 1.
    //
    // For example, _naf(7) will return an array b of 4 ints with:
    //     b[0] = -1
    //     b[1] =  0
    //     b[2] =  0
    //     b[3] =  1
    //
    int i = 0;
    int l = expi(k) + 2;
    int* naf = malloc(l * sizeof(int));

    while (cmpis(k, 0) > 0) {
        naf[i] = 0;
        if (_get_bit(k, 0)) {
            naf[i] = (int)(2 - (*int_LSW(k) & 3));
            k = subis(k, naf[i]);
        }
        k = shifti(k, -1);
        i++;
    }
    while (i < l) {
        naf[i] = 0;
        i++;
    }
    return naf;
}
//----------------------------------------------------------------------------
